import pytest
from pytest_socket import disable_socket

from .. import app


@pytest.fixture(scope='session', autouse=True)
def celeryconf():
    new_conf = dict(
        broker_url='memory://',
        result_backend='cache+memory://',
        gcn_bind_address='127.0.0.1',
        gcn_bind_port=53410,
        gcn_remote_address='127.0.0.1',
        task_always_eager=True,
        task_eager_propagates=True,
        lvalert_host='lvalert.invalid'
    )
    tmp = {key: app.conf[key] for key in new_conf.keys()}
    app.conf.update(new_conf)
    yield
    app.conf.update(tmp)


def pytest_runtest_setup():
    disable_socket()



from contextlib import contextmanager
import tempfile

__all__ = ('NamedTemporaryFile',)


@contextmanager
def NamedTemporaryFile(content=None, **kwargs):
    """Convenience wrapper for :func:`tempfile.NamedTemporaryFile` that writes
    some data to the file before handing it to the calling code.

    Parameters
    ----------

    content : str, bytes, None
        Initial contents of the file.

    \**kwargs
        Additional keyword arguments to pass to
        :func:`tempfile.NamedTemporaryFile`.
    """
    if isinstance(content, bytes):
        kwargs = dict(kwargs, mode='w+b')
    elif isinstance(content, str):
        kwargs = dict(kwargs, mode='w+')
    elif content is not None:
        raise TypeError('content is of unknown type')
    with tempfile.NamedTemporaryFile(**kwargs) as f:
        if content is not None:
            f.write(content)
            f.flush()
            f.seek(0)
        yield f



import io
from unittest.mock import patch

from glue.ligolw import utils
from glue.ligolw import lsctables
from glue.ligolw import table
from ligo.skymap.io.events.ligolw import ContentHandler
import pkg_resources
import pytest

from ..tasks.first2years import pick_coinc, upload_event

pytest.importorskip('lal')


def mock_now():
    import lal
    return lal.LIGOTimeGPS(1208808586)


@patch('lal.GPSTimeNow', mock_now)
def test_pick_coinc():
    coinc = pick_coinc()
    xmldoc, _ = utils.load_fileobj(io.BytesIO(coinc),
                                   contenthandler=ContentHandler)

    coinc_inspiral_table = table.get_table(
        xmldoc, lsctables.CoincInspiralTable.tableName)

    assert len(coinc_inspiral_table) == 1
    coinc_inspiral, = coinc_inspiral_table
    assert coinc_inspiral.get_end() <= mock_now()


@patch('lal.GPSTimeNow', mock_now)
@patch('gwcelery.tasks.gracedb.create_event', return_value='M1234')
@patch('gwcelery.tasks.gracedb.upload')
def test_upload_event(mock_upload, mock_create_event):
    coinc = pick_coinc()
    psd = pkg_resources.resource_filename(
        __name__, '../data/first2years/2016/psd.xml.gz')
    with open(psd, 'rb') as f:
        psd = f.read()

    upload_event()
    mock_create_event.assert_called_once_with(
        coinc, 'MDC', 'gstlal', 'CBC',
        'https://gracedb-playground.ligo.org/api/')
    mock_upload.assert_called_once_with(
        psd, 'psd.xml.gz', 'M1234', 'https://gracedb-playground.ligo.org/api/',
        'Noise PSD', ['psd'])



"""Create mock events from the "First Two Years" paper."""
import io

from celery.task import PeriodicTask
from celery.utils.log import get_task_logger
from glue.ligolw import utils
from glue.ligolw import lsctables
from glue.ligolw.table import get_table
import lal
from ligo.skymap.io.events.ligolw import ContentHandler
import numpy as np
import pkg_resources

from ..celery import app
from . import gracedb

log = get_task_logger(__name__)


@app.task(shared=False)
def pick_coinc():
    """Pick a coincidence from the "First Two Years" paper."""
    filename = pkg_resources.resource_filename(
        __name__, '../data/first2years/2016/gstlal.xml.gz')
    xmldoc = utils.load_filename(filename, contenthandler=ContentHandler)
    root, = xmldoc.childNodes

    # Remove unneeded tables
    for lsctable in (
            lsctables.FilterTable,
            lsctables.SegmentTable,
            lsctables.SegmentDefTable,
            lsctables.SimInspiralTable,
            lsctables.SummValueTable,
            lsctables.SearchSummVarsTable):
        root.removeChild(get_table(xmldoc, lsctable.tableName))

    coinc_inspiral_table = table = get_table(
        xmldoc, lsctables.CoincInspiralTable.tableName)

    # Determine event with most recent sideral time
    gps_time_now = lal.GPSTimeNow()
    gmsts = np.asarray([lal.GreenwichMeanSiderealTime(_.end) for _ in table])
    gmst_now = lal.GreenwichMeanSiderealTime(gps_time_now)
    div, rem = divmod(gmst_now - gmsts, 2 * np.pi)
    i = np.argmin(rem)
    new_gmst = div[i] * 2 * np.pi + gmsts[i]
    old_time = table[i].end
    new_time = lal.LIGOTimeGPS()
    result = lal.GreenwichMeanSiderealTimeToGPS(new_gmst, new_time)
    result.disown()
    del result
    delta_t = new_time - old_time
    target_coinc_event_id = int(table[i].coinc_event_id)

    # Remove unneeded rows
    table[:] = [row for row in table
                if int(row.coinc_event_id) == target_coinc_event_id]
    target_end_time = table[0].get_end()

    table = get_table(xmldoc, lsctables.CoincTable.tableName)
    table[:] = [row for row in table
                if int(row.coinc_event_id) == target_coinc_event_id]

    table = get_table(xmldoc, lsctables.CoincMapTable.tableName)
    table[:] = [row for row in table
                if int(row.coinc_event_id) == target_coinc_event_id]
    target_sngl_inspirals = frozenset(row.event_id for row in table)

    sngl_inspiral_table = table = get_table(
        xmldoc, lsctables.SnglInspiralTable.tableName)
    table[:] = [row for row in table if row.event_id in target_sngl_inspirals]

    table = get_table(xmldoc, lsctables.ProcessTable.tableName)
    table[:] = [row for row in table if row.program == 'gstlal_inspiral']
    target_process_ids = frozenset(row.process_id for row in table)

    table = get_table(xmldoc, lsctables.SearchSummaryTable.tableName)
    table[:] = [row for row in table if target_end_time in row.get_out()
                and row.process_id in target_process_ids]
    target_process_ids = frozenset(row.process_id for row in table)

    table = get_table(xmldoc, lsctables.ProcessTable.tableName)
    table[:] = [row for row in table if row.process_id in target_process_ids]

    table = get_table(xmldoc, lsctables.ProcessParamsTable.tableName)
    table[:] = [row for row in table if row.process_id in target_process_ids]

    # Shift event times
    for row in coinc_inspiral_table:
        row.end += delta_t
    for row in sngl_inspiral_table:
        row.end += delta_t
        row.end_time_gmst = lal.GreenwichMeanSiderealTime(row.end)

    coinc_xml = io.BytesIO()
    utils.write_fileobj(xmldoc, coinc_xml)
    return coinc_xml.getvalue()


@app.task(base=PeriodicTask, shared=False, run_every=720)
def upload_event(service='https://gracedb-playground.ligo.org/api/'):
    """Upload a random event from the "First Two Years" paper."""
    coinc = pick_coinc()
    psd = pkg_resources.resource_filename(
        __name__, '../data/first2years/2016/psd.xml.gz')
    with open(psd, 'rb') as f:
        psd = f.read()
    graceid = gracedb.create_event(coinc, 'MDC', 'gstlal', 'CBC', service)
    log.info('uploaded as %s', graceid)
    gracedb.upload(psd, 'psd.xml.gz', graceid, service, 'Noise PSD', ['psd'])



import os
import socket
from unittest.mock import patch
from xml.etree.ElementTree import XML

import pkg_resources
from pyxmpp2.exceptions import DNSError
import pytest

from ..tasks import lvalert
from .. import app


@pytest.fixture
def blacklist_cbc_gstlal_mdc():
    new_conf = dict(
        lvalert_node_whitelist={}
    )
    tmp = {key: app.conf[key] for key in new_conf.keys()}
    app.conf.update(new_conf)
    yield
    app.conf.update(tmp)


def test_filter_messages(blacklist_cbc_gstlal_mdc):
    xml = XML(pkg_resources.resource_string(__name__, 'data/lvalert_xmpp.xml'))
    messages = list(lvalert.filter_messages(xml))
    assert len(messages) == 0


@pytest.fixture
def whitelist_cbc_gstlal_mdc():
    new_conf = dict(
        lvalert_node_whitelist={'cbc_gstlal_mdc'}
    )
    tmp = {key: app.conf[key] for key in new_conf.keys()}
    app.conf.update(new_conf)
    yield
    app.conf.update(tmp)


def test_filter_messages_whitelist(whitelist_cbc_gstlal_mdc):
    xml = XML(pkg_resources.resource_string(__name__, 'data/lvalert_xmpp.xml'))
    messages = list(lvalert.filter_messages(xml))
    assert len(messages) == 1


@pytest.fixture
def netrc_lvalert2(tmpdir):
    path = str(tmpdir / 'netrc')
    with open(path, 'w') as f:
        print('machine', 'lvalert2.invalid', file=f)
        print('login', 'albert.einstein', file=f)
        print('password', 'foobar', file=f)
    with patch.dict(os.environ, NETRC=path):
        yield


def test_lvalert_netrc_does_not_match(netrc_lvalert2):
    """Test that we get the correct error message when the lvalert host is
    not in the netrc file."""
    with pytest.raises(RuntimeError, match='No matching netrc entry found'):
        lvalert.listen()


@pytest.fixture
def netrc_lvalert(tmpdir):
    path = str(tmpdir / 'netrc')
    with open(path, 'w') as f:
        print('machine', 'lvalert.invalid', file=f)
        print('login', 'albert.einstein', file=f)
        print('password', 'foobar', file=f)
    with patch.dict(os.environ, NETRC=path):
        yield


try:
    socket.gethostbyname('lvalert.invalid')
except socket.error:
    resolves_invalid_urls = False
else:
    resolves_invalid_urls = True


@pytest.mark.skipif(
    resolves_invalid_urls, reason='your DNS server is resolving invalid '
    'hostnames (maybe a home internet provider that provides a default '
    'landing page?)')
@pytest.mark.enable_socket
def test_lvalert_constructor(netrc_lvalert):
    """Test that we at least attempt to connect to a non-existent URL."""
    with pytest.raises(DNSError):
        lvalert.listen()



# import json
from unittest.mock import patch

import pkg_resources

from ..tasks.dispatch import dispatch


@patch('gwcelery.tasks.dispatch.annotate_fits')
@patch('gwcelery.tasks.dispatch.bayestar')
@patch('gwcelery.tasks.voevent.send.delay')
def test_dispatch_voevent(mock_send, mock_bayestar, mock_annotate_fits):
    """Test dispatch of a VOEvent message."""
    # Test LVAlert payload.
    payload = pkg_resources.resource_string(
        __name__, 'data/lvalert_voevent.json')

    # text = json.loads(payload)['object']['text']

    # Run function under test
    dispatch(payload)

    # Check that the correct tasks were dispatched.
    mock_annotate_fits.assert_not_called()
    mock_bayestar.assert_not_called()
    # FIXME: temporarily disable sending GCNs as per P. Brady request
    mock_send.assert_not_called()  # mock_send.assert_called_once_with(text)


@patch('gwcelery.tasks.dispatch.annotate_fits')
@patch('gwcelery.tasks.dispatch.bayestar')
def test_dispatch_label(mock_bayestar, mock_annotate_fits):
    """Test dispatch of a label message that should be ignored."""
    # Test LVAlert payload.
    payload = pkg_resources.resource_string(
        __name__, 'data/lvalert_label_dqv.json')

    # Run function under test
    dispatch(payload)

    # Check that no tasks were dispatched.
    mock_annotate_fits.assert_not_called()
    mock_bayestar.assert_not_called()


@patch('gwcelery.tasks.dispatch.annotate_fits')
@patch('gwcelery.tasks.dispatch.bayestar')
def test_dispatch_ignored(mock_bayestar, mock_annotate_fits):
    """Test dispatch of a detchar status message that should be ignored."""
    # Test LVAlert payload.
    payload = pkg_resources.resource_string(
        __name__, 'data/lvalert_detchar.json')

    # Run function under test
    dispatch(payload)

    # Check that no tasks were dispatched.
    mock_annotate_fits.assert_not_called()
    mock_bayestar.assert_not_called()


@patch('gwcelery.tasks.dispatch.bayestar')
def test_dispatch_psd(mock_bayestar):
    """Test dispatch of an LVAlert message for a PSD upload."""
    # Test LVAlert payload.
    payload = pkg_resources.resource_string(__name__, 'data/lvalert_psd.json')

    # Run function under test
    dispatch(payload)

    # Check that the correct tasks were dispatched.
    mock_bayestar.assert_called_once_with(
        'T250822', 'https://gracedb-test.ligo.org/api/')


@patch('gwcelery.tasks.dispatch.annotate_fits')
def test_dispatch_fits(mock_annotate_fits):
    """Test dispatch of an LVAlert message for a PSD upload."""
    # Test LVAlert payload.
    payload = pkg_resources.resource_string(__name__, 'data/lvalert_fits.json')

    # Run function under test
    dispatch(payload)

    # Check that the correct tasks were dispatched.
    mock_annotate_fits.assert_called_once_with(
        'bayestar.fits.gz,2', 'bayestar', 'T250822',
        'https://gracedb-test.ligo.org/api/', ['sky_loc'])



import os

import pytest

from ..util import NamedTemporaryFile


def test_named_temporary_file(tmpdir):
    """Test NamedTemporaryFile wrapper."""
    content = u'Hello world'
    with NamedTemporaryFile(prefix=str(tmpdir), content=content) as tmpfile:
        filename = tmpfile.name
        assert tmpfile.read() == content
    assert not os.path.exists(filename)


def test_named_temporary_file_unknown_type(tmpdir):
    """Test NamedTemporaryFile wrapper with content of the wrong type."""
    content = 12345
    with pytest.raises(TypeError):
        with NamedTemporaryFile(prefix=str(tmpdir), content=content):
            pass



"""Routing of LVAlert messages to other tasks."""
import json
import os

from ..celery import app
from .bayestar import bayestar
from .skymaps import annotate_fits
# from .voevent import send


@app.task(ignore_result=True, shared=False)
def dispatch(payload):
    """Parse an LVAlert message and dispatch it to other tasks."""
    # Parse JSON payload
    alert = json.loads(payload)

    # Determine GraceDB service URL
    try:
        try:
            self_link = alert['object']['links']['self']
        except KeyError:
            self_link = alert['object']['self']
    except KeyError:
        # Cannot deduce API link
        return
    base, api, _ = self_link.partition('/api/')
    service = base + api

    # Determine GraceDB ID
    graceid = alert['uid']

    if alert['alert_type'] == 'update' and 'voevent_type' in alert['object']:
        # FIXME: temporarily disable sending GCNs as per P. Brady request
        pass  # send.delay(alert['object']['text'])
    elif alert['alert_type'] == 'update' and alert.get('file'):
        _, versioned_filename = os.path.split(alert['object']['file'])
        filename, _, _ = versioned_filename.rpartition(',')
        filebase, fitsext, _ = filename.rpartition('.fits')
        tags = alert['object']['tag_names']
        if fitsext:
            annotate_fits(
                versioned_filename, filebase, graceid, service, tags).delay()
        elif filename == 'psd.xml.gz':
            bayestar(graceid, service).delay()



"""Basic single-endpoint VOEvent broker."""
import socket
import struct
from urllib.parse import urlparse, urlunparse

from celery import Task
from celery.utils.log import get_task_logger
from celery_eternal import EternalProcessTask
import gcn
import lxml.etree

from ..celery import app
from ..tasks import gracedb

# Logging
log = get_task_logger(__name__)

_size_struct = struct.Struct("!I")


class SendTask(Task):

    def __init__(self):
        self.conn = None


@app.task(queue='voevent', base=SendTask, ignore_result=True, bind=True,
          autoretry_for=(socket.error,), default_retry_delay=0.001,
          retry_backoff=True, retry_kwargs=dict(max_retries=None),
          shared=False)
def send(self, payload):
    """Task to send VOEvents. Supports only a single client."""
    payload = payload.encode('utf-8')
    nbytes = len(payload)

    conn = self.conn
    self.conn = None

    if conn is None:
        log.info('creating new socket')
        sock = socket.socket(socket.AF_INET)
        try:
            sock.bind((app.conf['gcn_bind_address'],
                       app.conf['gcn_bind_port']))
            sock.listen(0)
            while True:
                conn, (addr, _) = sock.accept()
                if addr == app.conf['gcn_remote_address']:
                    break
                else:
                    log.error('connection denied to remote host %s', addr)
        finally:
            sock.close()
        conn.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER,
                        struct.pack('ii', 1, 0))
        conn.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1)

    log.info('sending payload of %d bytes', nbytes)
    try:
        conn.sendall(_size_struct.pack(nbytes) + payload)
    except:  # noqa
        try:
            conn.shutdown(socket.SHUT_RDWR)
        except:  # noqa
            log.exception('failed to shut down socket')
        conn.close()
        raise
    self.conn = conn


@app.task(ignore_result=True, shared=False)
def validate(payload):
    """Check that the contents of a public LIGO/Virgo GCN matches the original
    VOEvent in GraceDB."""
    root = lxml.etree.fromstring(payload)

    # Which GraceDB ID does this refer to?
    graceid = root.find("./What/Param[@name='GraceID']").attrib['value']

    # Which VOEvent does this refer to?
    u = urlparse(root.attrib['ivorn'])
    assert u.scheme == 'ivo', (
        'IVORN has unexpected scheme: {!r}'.format(u.scheme))
    assert u.netloc == 'nasa.gsfc.gcn', (
        'IVORN has unexpected netloc: {!r}'.format(u.netloc))
    assert u.path == '/LVC', (
        'IVORN has unexpected path: {!r}'.format(u.path))
    local_id = u.fragment
    filename = local_id + '.xml'

    # Which GraceDB server does this refer to?
    u = urlparse(root.find("./What/Param[@name='EventPage']").attrib['value'])
    service = urlunparse((u.scheme, u.netloc, '/api/', None, None, None))

    # Download and parse original VOEvent
    orig = lxml.etree.fromstring(gracedb.download(filename, graceid, service))

    xpath = ".//Param[@name='{}']"
    for orig_name, root_name in [
            ['skymap_fits_shib', 'SKYMAP_URL_FITS_SHIB'],
            ['skymap_fits_x509', 'SKYMAP_URL_FITS_X509'],
            ['skymap_fits_basic', 'SKYMAP_URL_FITS_BASIC'],
            ['skymap_png_shib', 'SKYMAP_URL_PNG_SHIB'],
            ['skymap_png_x509', 'SKYMAP_URL_PNG_X509'],
            ['skymap_png_basic', 'SKYMAP_URL_PNG_BASIC']]:

        orig_elem = orig.find(xpath.format(orig_name))
        root_elem = root.find(xpath.format(root_name))

        if orig_elem is None:
            assert root_elem is None, (
                'GCN has unexpected parameter: {!r}'.format(root_name))
        else:
            assert root_elem is not None, (
                'GCN is missing parameter: {!r}'.format(root_name))
            orig_value = orig_elem.attrib.get('value')
            root_value = root_elem.attrib.get('value')
            assert root_value == orig_value, (
                'GCN parameter {!r} has value {!r}, but '
                'original VOEvent parameter {!r} '
                'has value {!r}'.format(
                    root_name, root_value, orig_name, orig_value))

    # Find matching GraceDB log entry
    log = gracedb.get_log(graceid, service)
    entry, = (e for e in log if e['filename'] == filename)
    log_number = entry['N']

    # Tag the VOEvent to indicate that it was received correctly
    gracedb.create_tag('gcn_received', log_number, graceid, service)


@gcn.include_notice_types(
    gcn.notice_types.LVC_PRELIMINARY,
    gcn.notice_types.LVC_INITIAL,
    gcn.notice_types.LVC_UPDATE
)
def handle(payload, root):
    validate.delay(payload)


@app.task(base=EternalProcessTask, shared=False)
def listen():
    """Listen for public GCNs and validate the contents of public LIGO/Virgo
    GCNs by passing their contents to :obj:`validate`."""
    gcn.listen(handler=handle)



from unittest.mock import patch
import pkg_resources
import pytest
from xml.sax import SAXParseException

from ..tasks.bayestar import bayestar, localize


def mock_download(filename, graceid, service):
    assert graceid == 'T12345'
    assert service == 'https://gracedb.invalid/api/'
    if filename == 'coinc.xml':
        return pkg_resources.resource_string(__name__, 'coinc.xml')
    elif filename == 'psd.xml.gz':
        return pkg_resources.resource_string(__name__, 'psd.xml.gz')
    else:
        raise ValueError


@patch('gwcelery.tasks.gracedb.download', mock_download)
@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_bayestar(mock_gracedb):
    # Run function under test
    bayestar('T12345', 'https://gracedb.invalid/api/')


@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_localize_bad_psd(mock_gracedb):
    """Test running BAYESTAR with a pad PSD file"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = b''

    # Run function under test
    with pytest.raises(SAXParseException):
        localize(
            (coinc, psd), 'G211117', 'https://gracedb.invalid/api/')


@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_localize(mock_gracedb):
    """Test running BAYESTAR on G211117"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = pkg_resources.resource_string(__name__, 'data/psd.xml.gz')

    # Run function under test
    fitscontent = localize(
        (coinc, psd), 'G211117', 'https://gracedb.invalid/api/')

    # FIXME: should do some sanity checks of the sky map here
    assert fitscontent


@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_localize_detector_disabled(mock_gracedb):
    """Test running BAYESTAR on G211117 with L1 disabled"""
    # Test data
    coinc = pkg_resources.resource_string(__name__, 'data/coinc.xml')
    psd = pkg_resources.resource_string(__name__, 'data/psd.xml.gz')

    # Run function under test
    fitscontent = localize(
        (coinc, psd), 'G211117', 'https://gracedb.invalid/api/',
        disabled_detectors=['L1'])

    # FIXME: should do some sanity checks of the sky map here
    assert fitscontent



import json
import socket
from threading import Thread
from time import sleep
from unittest.mock import MagicMock

from gcn.voeventclient import _recv_packet
import lxml.etree
import pkg_resources
import pytest

from ..tasks.voevent import handle, listen, send
from .. import app

# Test data
with pkg_resources.resource_stream(
        __name__, 'data/lvalert_voevent.json') as f:
    lvalert = json.load(f)
voevent = lvalert['object']['text']


@pytest.fixture
def send_thread():
    thread = Thread(target=send, args=(voevent,))
    thread.daemon = True
    thread.start()
    sleep(1)
    yield thread
    thread.join()
    if send.conn is not None:
        try:
            send.conn.close()
        except socket.error:
            pass
        send.conn = None


@pytest.mark.enable_socket
def test_send_connection_closed(send_thread):
    """Test sending a VOEvent over loopback to a connection that
    is immediately closed."""
    sock = socket.socket(socket.AF_INET)
    sock.connect(('127.0.0.1', 53410))
    sock.shutdown(socket.SHUT_RDWR)
    sock.close()


@pytest.mark.enable_socket
def test_send(send_thread):
    """Test sending a VOEvent over loopback."""
    # First, simulate connecting from a disallowed IP address.
    # The connection should be refused.
    app.conf['gcn_remote_address'] = '192.0.2.0'
    sock = socket.socket(socket.AF_INET)
    try:
        sock.settimeout(0.1)
        with pytest.raises(socket.error):
            sock.connect(('127.0.0.1', 53410))
            packet = _recv_packet(sock)
    finally:
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        sock.close()

    # Now, simulate connecting from the allowed IP address.
    # The VOEvent should be received.
    app.conf['gcn_remote_address'] = '127.0.0.1'
    sock = socket.socket(socket.AF_INET)
    try:
        sock.settimeout(0.1)
        sock.connect(('127.0.0.1', 53410))
        packet = _recv_packet(sock)
    finally:
        try:
            sock.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass
        sock.close()
    assert packet == voevent.encode('utf-8')


@pytest.fixture
def fake_gcn(celeryconf, monkeypatch):

    def mock_download(filename, graceid, service):
        assert filename == 'G298048-1-Initial.xml'
        assert graceid == 'G298048'
        assert service == 'https://gracedb.ligo.org/api/'
        return pkg_resources.resource_string(__name__, 'data/' + filename)

    def mock_get_log(graceid, service):
        assert graceid == 'G298048'
        assert service == 'https://gracedb.ligo.org/api/'
        return json.loads(
            pkg_resources.resource_string(__name__, 'data/G298048_log.json'))

    def mock_create_tag(tag, n, graceid, service):
        assert tag == 'gcn_received'
        assert n == 532
        assert graceid == 'G298048'
        assert service == 'https://gracedb.ligo.org/api/'

    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.download', mock_download)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.get_log', mock_get_log)
    monkeypatch.setattr(
        'gwcelery.tasks.gracedb.create_tag', mock_create_tag)

    # Check the real GCN notice, which is valid.
    payload = pkg_resources.resource_string(
        __name__, 'data/G298048-1-Initial.gcn.xml')
    root = lxml.etree.fromstring(payload)
    yield root


@pytest.mark.enable_socket
def test_listen(monkeypatch):
    """Test that the listen task would correctly launch gcn.listen()."""
    mock_gcn_listen = MagicMock()
    monkeypatch.setattr('gcn.listen', mock_gcn_listen)
    listen.run()
    mock_gcn_listen.assert_called_once_with(handler=handle)


def test_handle(fake_gcn):
    """Test that the fake GCN notice matches what we actually sent."""
    handle(lxml.etree.tostring(fake_gcn), fake_gcn)


def test_handle_mismatched_param(fake_gcn):
    """Test that we correctly detect mismatched parameter values in GCNs."""
    xpath = ".//Param[@name='SKYMAP_URL_FITS_BASIC']"
    fake_gcn.find(xpath).attrib['value'] = ('wrong, you fool!')
    with pytest.raises(
            AssertionError,
            match="^GCN parameter 'SKYMAP_URL_FITS_BASIC' has value 'wrong, "
            "you fool!', but original VOEvent parameter 'skymap_fits_basic' "
            "has value 'https://gracedb.ligo.org/apibasic/events/G298048/"
            "files/bayestar.fits.gz'$"):
        handle(lxml.etree.tostring(fake_gcn), fake_gcn)


def test_handle_unexpected_param(fake_gcn, monkeypatch):
    """Test that we correctly detect unexpected paremeters in GCNs."""

    def mock_download(filename, graceid, service):
        assert filename == 'G298048-1-Initial.xml'
        assert graceid == 'G298048'
        assert service == 'https://gracedb.ligo.org/api/'
        payload = pkg_resources.resource_string(__name__, 'data/' + filename)
        root = lxml.etree.fromstring(payload)
        elem = root.find(".//Group[@type='GW_SKYMAP']")
        elem.getparent().remove(elem)
        return lxml.etree.tostring(root)

    monkeypatch.setattr('gwcelery.tasks.gracedb.download', mock_download)

    with pytest.raises(
            AssertionError,
            match="^GCN has unexpected parameter: 'SKYMAP_URL_FITS_[A-Z]*'$"):
        handle(lxml.etree.tostring(fake_gcn), fake_gcn)


def test_handle_missing_param(fake_gcn, monkeypatch):
    """Test that we correctly detect unexpected paremeters in GCNs."""
    elem = fake_gcn.find(".//Param[@name='SKYMAP_URL_FITS_BASIC']")
    elem.getparent().remove(elem)
    with pytest.raises(
            AssertionError,
            match="^GCN is missing parameter: 'SKYMAP_URL_FITS_BASIC'$"):
        handle(lxml.etree.tostring(fake_gcn), fake_gcn)


def test_handle_wrong_ivorn_path(fake_gcn):
    """Test that we correctly detect the wrong IVORN path."""
    fake_gcn.attrib['ivorn'] = 'ivo://nasa.gsfc.gcn/wrong#G298048-1-Initial'
    with pytest.raises(
            AssertionError, match="^IVORN has unexpected path: '/wrong'$"):
        handle(lxml.etree.tostring(fake_gcn), fake_gcn)


def test_handle_wrong_ivorn_netloc(fake_gcn):
    """Test that we correctly detect the wrong IVORN netloc."""
    fake_gcn.attrib['ivorn'] = 'ivo://wrong/LVC#G298048-1-Initial'
    with pytest.raises(
            AssertionError, match="^IVORN has unexpected netloc: 'wrong'$"):
        handle(lxml.etree.tostring(fake_gcn), fake_gcn)


def test_handle_wrong_ivorn_scheme(fake_gcn):
    """Test that we correctly detect the wrong IVORN scheme."""
    fake_gcn.attrib['ivorn'] = 'wrong://nasa.gsfc.gcn/LVC#G298048-1-Initial'
    with pytest.raises(
            AssertionError, match="^IVORN has unexpected scheme: 'wrong'$"):
        handle(lxml.etree.tostring(fake_gcn), fake_gcn)



from unittest.mock import patch

from ..tasks import gracedb


def test_create_event(monkeypatch):

    class MockResponse(object):

        def json(self):
            return {'graceid': 'T12345'}

    class MockGraceDb(object):

        def __init__(self, service):
            assert service == 'service'

        def createEvent(self, group, pipeline, filename, search=None,
                        labels=None, offline=False, filecontents=None,
                        **kwargs):
            assert group == 'group'
            assert pipeline == 'pipeline'
            assert filename == 'initial.data'
            assert search == 'search'
            assert filecontents == 'filecontents'
            return MockResponse()

    monkeypatch.setattr('ligo.gracedb.rest.GraceDb', MockGraceDb)

    graceid = gracedb.create_event('filecontents', 'search', 'pipeline',
                                   'group', 'service')
    assert graceid == 'T12345'


@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_create_tag(mock_gracedb):
    # Run function under test.
    gracedb.create_tag('tag', 'n', 'graceid', 'service')

    # Check that GraceDb was instantiated once.
    mock_gracedb.assert_called_once_with('service')

    # Check that one file was downloaded.
    mock_gracedb.return_value.createTag.assert_called_once_with(
        'graceid', 'n', 'tag')


@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_download(mock_gracedb):
    # Run function under test.
    gracedb.download('filename', 'graceid', 'service')

    # Check that GraceDb was instantiated once.
    mock_gracedb.assert_called_once_with('service')

    # Check that one file was downloaded.
    mock_gracedb.return_value.files.assert_called_once_with(
        'graceid', 'filename', raw=True)


def test_get_log(monkeypatch):

    class logs(object):

        def json(self):
            return {'log': 'stuff'}

    class MockGraceDb(object):

        def __init__(self, service):
            assert service == 'service'

        def logs(self, graceid):
            assert graceid == 'graceid'
            return logs()

    monkeypatch.setattr('ligo.gracedb.rest.GraceDb', MockGraceDb)

    # Run function under test.
    ret = gracedb.get_log('graceid', 'service')

    assert ret == 'stuff'


@patch('ligo.gracedb.rest.GraceDb', autospec=True)
def test_upload(mock_gracedb):
    # Run function under test.
    gracedb.upload(
        'filecontents', 'filename', 'graceid', 'service', 'message', 'tags')

    # Check that GraceDb was instantiated once.
    mock_gracedb.assert_called_once_with('service')

    # Check that one file was uploaded.
    mock_gracedb.return_value.writeLog.assert_called_once_with(
        'graceid', 'message', 'filename', 'filecontents', 'tags')



"""LVAlert client."""
import netrc
import os
import uuid

from celery_eternal import EternalTask
from celery.utils.log import get_task_logger
from pyxmpp2.client import Client
from pyxmpp2.interfaces import (EventHandler, event_handler,
                                message_stanza_handler, QUIT,
                                XMPPFeatureHandler)
from pyxmpp2.jid import JID
from pyxmpp2.mainloop.interfaces import TimeoutHandler, timeout_handler
from pyxmpp2.settings import XMPPSettings
from pyxmpp2.streamevents import DisconnectedEvent

from ..celery import app
from .dispatch import dispatch

# Logging
log = get_task_logger(__name__)

ns = {'ns1': 'http://jabber.org/protocol/pubsub#event',
      'ns2': 'http://jabber.org/protocol/pubsub'}


def filter_messages(xml):
    for node in xml.iterfind('.//ns1:items[@node]', ns):
        log.info('LVAlert node: %r', node.attrib['node'])
        if node.attrib['node'] in app.conf['lvalert_node_whitelist']:
            for entry in node.iterfind('.//ns2:entry', ns):
                yield entry.text
        else:
            log.info(
                'ignorning because LVAlert node %r is not in whitelist %r',
                node.attrib['node'], app.conf['lvalert_node_whitelist'])


class LVAlertClient(EventHandler, TimeoutHandler, XMPPFeatureHandler):

    def __init__(self, server, task=None):
        # Look up username and password from the netrc file.
        netrcfile = os.environ.get('NETRC')
        auth = netrc.netrc(netrcfile).authenticators(server)
        if auth is None:
            raise RuntimeError('No matching netrc entry found')
        username, _, password = auth

        # Create a JID with a unique resource name.
        resource = uuid.uuid4().hex
        jid = JID(username + '@' + server + '/' + resource)

        settings = XMPPSettings(dict(
            starttls=True, tls_verify_peer=False, password=password))

        self.__task = task
        self.__client = Client(jid, [self], settings)
        self.__client.main_loop.add_handler(self)

    @event_handler(DisconnectedEvent)
    def __handle_disconnected(self, *args, **kwargs):
        return QUIT

    @message_stanza_handler()
    def __handle_message(self, stanza):
        log.info('Got message')
        for text in filter_messages(stanza.as_xml()):
            log.debug('Dispatching: %s', text)
            dispatch.s(text).delay()
        return True

    @timeout_handler(1, recurring=True)
    def __handle_timeout(self, *args, **kargs):
        if self.__task is not None and self.__task.is_aborted():
            log.info('Task aborted, quitting event loop')
            self.__client.disconnect()
        return True

    def run(self):
        log.info('Connecting to %r', str(self.__client.jid))
        self.__client.connect()
        self.__client.run()
        log.info('Reached end of main loop')


@app.task(base=EternalTask, bind=True, shared=False)
def listen(self):
    """Listen for LVAlert messages forever. Each message is processed by
    passing it to :func:`~gwcelery.tasks.dispatch.dispatch`."""
    LVAlertClient(app.conf['lvalert_host'], self).run()



from celery import Celery

# Celery application object.
# Use pickle serializer, because it supports byte values.
# Use redis broker, because it supports locks (and thus singleton tasks).
app = Celery(
    'gwcelery', broker='redis://',
    config_source=dict(
        accept_content=['json', 'pickle'],
        event_serializer='json',
        result_serializer='pickle',
        task_serializer='pickle',
        lvalert_host='lvalert-test.cgca.uwm.edu',
        lvalert_node_whitelist={'cbc_gstlal', 'cbc_pycbc', 'cbc_mbtaonline',
                                'cbc_gstlal_mdc'},
        gcn_bind_address='',
        gcn_bind_port=5341,
        gcn_remote_address='128.183.96.236'  # capella2.gsfc.nasa.gov
    )
)

# Use the same URL for both the result backend and the broker.
app.conf['result_backend'] = app.conf.broker_url



"""Communication with GraceDB."""
from ligo.gracedb import rest

from ..celery import app


@app.task(shared=False)
def create_event(filecontents, search, pipeline, group, service):
    client = rest.GraceDb(service)
    response = client.createEvent(group=group, pipeline=pipeline,
                                  filename='initial.data', search=search,
                                  filecontents=filecontents)
    return response.json()['graceid']


@app.task(ignore_result=True, shared=False)
def create_tag(tag, n, graceid, service):
    rest.GraceDb(service).createTag(graceid, n, tag)


@app.task(shared=False)
def download(filename, graceid, service):
    """Download a file from GraceDB."""
    return rest.GraceDb(service).files(graceid, filename, raw=True).read()


@app.task(shared=False)
def get_log(graceid, service):
    return rest.GraceDb(service).logs(graceid).json()['log']


@app.task(ignore_result=True, shared=False)
def upload(filecontents, filename, graceid, service, message, tags):
    """Upload a file to GraceDB."""
    rest.GraceDb(service).writeLog(
        graceid, message, filename, filecontents, tags)



"""All Celery tasks are declared in submodules of this module."""
import importlib
import os
import pkgutil

# Import all submodules of this module
modules = vars()
for _, module, _ in pkgutil.iter_modules([os.path.dirname(__file__)]):
    modules[module] = importlib.import_module('.' + module, __name__)
    del module

# Clean up
del importlib, modules, os, pkgutil
